from __future__ import with_statement # requires Python 2.5
import os
from collections import defaultdict
import re

import struct
import ssconst
from ssspeechmap import SpeechMap, SoundSource, SoundEntry
import sstextdata
import ssmisc
from ssutil import readUint32BE

class TextImporter(object):
    def _parse_text_data(self, text):
        raise NotImplementedException()

    def import_from_resource(self, path):
        """ Returns a text data object, stripped of speech codes, and
        a map of sound IDs (actually offsets) to line numbers.
        """
        text_data = sstextdata.TextData()
        with file(path, 'r') as text_file:
            sound_map = defaultdict(list)
            for i, text in enumerate(text_file):
                text, sound_id = self._parse_text_data(text)
                if sound_id is not None:
                    sound_map[sound_id].append(i)
                text_data.append(text)
        return (text_data, sound_map)

class TextMonsterImporter(TextImporter):
    # Match text like the following string:
    # \xFF\x0A\x18\x7B\xFF\x0A\x09\x00\xFF\x0A\x0A\x00\xFF\x0A\x00\x00
    # Group the address and tag values
    sound_regex = re.compile(('\\\\xFF\\\\x0A' +
                    '\\\\x([0-9A-Fa-f]{2})\\\\x([0-9A-Fa-f]{2})' ) * 4 +
                    '(?P<text>.*)')

    def _parse_text_data(self, text):
        """ Returns a 2-item tuple, containing:
         - text data
         - sound ID (offset to the speech data)

        If the line of text does not have a sound, the sound ID will be "None".
        """
        match = self.sound_regex.match(text)
        if match is None:
            return (text, None)
        assert len(match.groups()) == 9
        # Convert from text, big-endian, hex, into int.
        sound_id = int(''.join(match.group(4, 3, 2, 1)), 16)
        sound_id = str(sound_id) # back to a string
        # Commented out - we don't care about the aux data size, so just ignore it.
        #aux_data_size = int(''.join(match.group(8, 7, 6, 5)), 16)
        text = match.group('text')
        return (text, sound_id)

class TextBunImporter(TextImporter):
    # Match text like the following string:
    # /LOGO.064/This is my Pen Ultimate personal digital assistant.
    # Group the ID (LOGO.064) and text.
    sound_regex = re.compile(
                    '/(?P<sound_id>.{1,8}\\.[0-9]{3})/' +
                    '(?P<text>.*)')

    def _parse_text_data(self, text):
        """ Returns a 2-item tuple, containing:
         - text data
         - sound ID (offset to the speech data)

        If the line of text does not have a sound, the sound ID will be "None",
        and the line of text will be returned unchanged.
        """
        match = self.sound_regex.match(text)
        if match is None:
            return (text, None)
        # Return the ID.
        sound_id = match.group('sound_id')
        text = match.group('text')
        return (text, sound_id)

class SoundResourceImporter(object):
    resource_type = None

    def __init__(self):
        self.compression_type = None

    def import_from_resource(self, path):
        """ Returns a list of SoundEntry objects."""
        raise NotImplementedException()

    def determine_compression_type(self, res_path):
        raise NotImplementedException()

class MonsterImporter(SoundResourceImporter):
    resource_type = ssconst.TYPE_MONSTER

class OriginalMonsterImporter(MonsterImporter):
    def determine_compression_type(self, res_path):
        return ssconst.COMP_ORIGINAL

class CompressedMonsterImporter(MonsterImporter):
    compressed_monster_types = {
        '.sog' : ssconst.COMP_OGG,
        '.so3' : ssconst.COMP_MP3,
        '.sof' : ssconst.COMP_FLAC
    }

    def determine_compression_type(self, res_path):
        res_ext = os.path.splitext(res_path)[1].lower()
        if res_ext in self.compressed_monster_types:
            return self.compressed_monster_types[res_ext]
        raise ssmisc.SSException(("Unrecognised file extension for compressed " +
        "MONSTER resource:\n" +
        "%s.\n" +
        "Expected one of %s") %
        (res_path, str(self.compressed_monster_types.keys())))

    def import_from_resource(self, path):
        sounds = []

        self.compression_type = self.determine_compression_type(path)
        
        with file(path, 'rb') as res_file:
            (i_index_size,) = struct.unpack('>I', res_file.read(4))

            end_table_offset = i_index_size + 4

            while res_file.tell() < end_table_offset:
                # orig_off = original offset that the game looks up.
                # new_off = offset in the compressed monster file, after the index table
                # tag_size = total size of lip-synching tags (each tag is 16-bit BE)
                # sound_size = size of the individual compressed voice file
                orig_off, new_off, tag_size, sound_size = struct.unpack('>4I', res_file.read(16))

                # go read the lip synch tags as metadata
                current_loc = res_file.tell()
                res_file.seek(end_table_offset + new_off)
                metadata = {
                    "LipSynch" : struct.unpack('>' + str(tag_size >> 1) + 'H', res_file.read(tag_size))
                }
                res_file.seek(current_loc)

                source = SoundSource(
                    path,
                    end_table_offset + new_off + tag_size,
                    sound_size
                )
                sound = SoundEntry(
                   str(orig_off), # ID
                   source, # where we should read sound data from
                   metadata # lip synch tags
                )
                sounds.append(sound)

        return (sounds, self.resource_type, self.compression_type)

class BunImporter(SoundResourceImporter):
    resource_type = ssconst.TYPE_BUN

class OriginalBunImporter(BunImporter):
    def determine_compression_type(self, res_path):
        return ssconst.COMP_ORIGINAL

class CompressedBunImporter(BunImporter):
    """
    Compressed BUN format is as follows. All values are 32-bit BE unsigned ints,
    unless otherwise stated.

    "LB23"
    offset of the cBundleTable
    number of entries in the cBundleTable

    for each sound:
    - "RMAP", followed by:
    -- 3 (version)
    -- 16 (bits)
    -- frequency
    -- channels
    -- numRegions
    -- numJumps
    -- numSyncs
    -- numMarkers
    -- for i in numRegions:
    --- offset
    --- length
    -- for i in numJumps:
    --- offset
    --- dest
    --- hook ID
    --- fade delay
    -- for i in numSyncs:
    --- size of the sync data
    --- the sync data (variable size)
    -- for i in numMarkers:
    --- pos
    --- length
    --- the marker data (variable size)
    - sound data

    for each entry in table:
    - filename (24 characters)
    - offset of the file in the bundle
    - size of the file in the bundle

    Generally, at least for speech in The Dig, there will be entries for
    an RMAP file, and at least 1 sound file. Sound files are named per
    region. For instance, sound "AIRLOCK.001" in a compressed file has:
    - AIRLOCK.001.map
    - AIRLOCK.001_reg000.ogg
    The "_reg000" represents the region.

    (I don't think ScummSpeaks will support multiple regions, since regions are
    used more for music than speech. The Dig speech does not have any regions
    other than 000.)
    """

    compressed_bun_types = {
        '.ogg' : ssconst.COMP_OGG,
        '.mp3' : ssconst.COMP_MP3,
        '.fla' : ssconst.COMP_FLAC,
        '.flac' : ssconst.COMP_FLAC,
    }

    name_key_regex = re.compile(
                    '(?P<name_key>.{1,8}\\.[0-9]{3})' +
                    '(?P<extension>.*)')

    def _get_name_key(self, name):
        match = self.name_key_regex.match(name)
        if match is None:
            raise ssmisc.SSException("Unexpected format for sound file name: %s" % name)
        return match.group('name_key')

    def _parse_metadata(self, rmap, sound_size):
        # TODO?
        metadata = dict()
        metadata.update(rmap)
        return metadata

    def determine_compression_type(self, res_path):
        # Look through the file names stored in the bundle until we find one
        #  with an extension that we recognise (e.g. "AIRLOCK.001_res000.ogg").
        with file(res_path, 'rb') as res_file:
            assert res_file.read(4) == "LB23"
            table_offset, table_num_entries = readUint32BE(res_file, 2)
            res_file.seek(table_offset)
            for i in xrange(table_num_entries):
                name = res_file.read(24).partition('\x00')[0]
                ext = os.path.splitext(name)[1].lower()
                if ext in self.compressed_bun_types:
                    return self.compressed_bun_types[ext]
                res_file.seek(8, os.SEEK_CUR) # skip the size and offset data
        raise ssmisc.SSException("Could not determine compression type of " +
            "compressed BUN resource:\n%s" % res_path)

        res_ext = os.path.splitext(res_path)[1].lower()
        if res_ext in self.compressed_monster_types:
            return self.compressed_monster_types[res_ext]
        raise ssmisc.SSException("Unrecognised file extension for compressed " +
        "MONSTER resource:\n" +
        "%s.\n" +
        "Expected one of %s" %
        (res_path, str(self.compressed_monster_types.keys())))

    def import_from_resource(self, path):
        sounds = []

        self.compression_type = self.determine_compression_type(path)

        with file(path, 'rb') as res_file:
            assert res_file.read(4) == "LB23"
            table_offset, table_num_entries = readUint32BE(res_file, 2)

            # Skip to the table data and read it in.
            # Store each entry in a dict, key is the sound
            #  name (as used by the game scripts), value
            #  is another dict with two keys, "RMAP" and "Sound".
            res_file.seek(table_offset)
            table_data = defaultdict(dict)
            for i in xrange(table_num_entries):
                # Split on first "\x00" char rather than strip trailing "\x00"
                #  chars, as I've noticed the ScummVM tool seems to add junk
                #  data. See the entry for "AST2.031.rmap" for The Dig,
                #  which adds the following bytes:
                #   00 00 00 00 1E 50 04 00 90 00 37 00
                name = res_file.read(24).partition('\x00')[0]
                offset, size = readUint32BE(res_file, 2)
                entry = (name, offset, size)
                name_key = self._get_name_key(name)
                data_key = "RMAP" if name.lower().endswith('.map') else "Sound"
                table_data[name_key][data_key] = entry

            # Read all RMAP data
            for sound_id, data in table_data.items():
                rmap_file_name, rmap_offset, rmap_size = data['RMAP']
                sound_file_name, sound_offset, sound_size = data['Sound']

                res_file.seek(rmap_offset)
                tag_name = res_file.read(4)
                assert tag_name == "RMAP"
                version, bit_depth, frequency, channels, num_regions, \
                num_jumps, num_syncs, num_markers = readUint32BE(res_file, 8)
                assert version == 3 # version
                assert bit_depth == 16 # bit depth
                regions = []
                jumps = []
                syncs = []
                markers = []
                for j in xrange(num_regions):
                    offset, length = readUint32BE(res_file, 2)
                    regions.append(
                        dict(Offset=offset, Length=length)
                    )
                # Jumps, Syncs, and Markers are empty for all speech in The Dig,
                #  but support them anyway, just in case.
                for j in xrange(num_jumps):
                    offset, dest, hook_id, fade_delay = readUint32BE(res_file, 4)
                    jumps.append(
                        dict(Offset=offset, Destination=dest, HookId=hook_id, FadeDelay=fade_delay)
                    )
                for j in xrange(num_syncs):
                    size = readUint32BE(res_file)
                    sync_data = res_file.read(size)
                    syncs.append(
                        dict(SyncData=sync_data)
                    )
                for j in xrange(num_markers):
                    pos, length = readUint32BE(res_file, 2)
                    marker_data = res_file.read(size)
                    markers.append(
                        dict(Position=pos, MarkerData=marker_data)
                    )
                # Paranoia check
                assert res_file.tell() == rmap_offset + rmap_size
               
                # Convert RMAP data into metadata
                rmap = dict(
                    Version=version,
                    BitDepth=bit_depth,
                    Frequency=frequency,
                    Channels=channels,
                    Regions=regions,
                    Jumps=jumps,
                    Syncs=syncs,
                    Markers=markers
                )
                metadata = self._parse_metadata(rmap, sound_size)

                source = SoundSource(
                    path,
                    sound_offset,
                    sound_size
                )
                sound = SoundEntry(
                   sound_id, # ID
                   source, # where we should read sound data from
                   metadata # lip synch tags
                )
                sounds.append(sound)

        # Sort by ID
        sounds.sort(key=(lambda s: s.id))
        return (sounds, self.resource_type, self.compression_type)

class SoundResourceImporterFactory(object):
    compressed_monster_types = frozenset([
        '.sog',
        '.so3',
        '.sof'
    ])

    def __new__(cls, res_path, *args, **kwds):
        """Returns a sound importer object based on the input file."""
        res_ext = os.path.splitext(res_path)[1].lower()
        if res_ext == '.sou':
            return None # TODO: support original MONSTER files
        elif res_ext in cls.compressed_monster_types:
            return CompressedMonsterImporter()
        elif res_ext == '.bun':
            with file(res_path, 'rb') as res_file:
                tag_name = res_file.read(4)
            if tag_name == 'LB83':
                return None # TODO support original BUN files.
            elif tag_name == 'LB23':
                return CompressedBunImporter()
            else:
                return None

class TextImporterFactory(object):
    def __new__(cls, res_type, *args, **kwds):
        if res_type == ssconst.TYPE_MONSTER:
            return TextMonsterImporter()
        elif res_type == ssconst.TYPE_BUN:
            return TextBunImporter()